#include <gpl.h>
#include <l-portable/icon-side.xpm>

//standard
#include <cstdlib>
#include <cstdio>	//too used to printf
#include <vector>
#include <cmath>
#include <cstring>
#include <iostream>
#include <inttypes.h>

//system
#include <unistd.h>
#include <sys/types.h>	//for getpid: http://linux.die.net/man/2/getpid
#include <libgen.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <errno.h>
#include <sys/wait.h>

//gtk
#include <gtk/gtk.h>
#include <gtk/gtkimagemenuitem.h>

//gvfs/gio
#include <gio/gio.h>

//xml
#include <tinyxml/tinyxml.h>

//curl
#include <curl/curl.h>

//regex
#include <pcre.h>

//bz2
#include <bzlib.h>

//tar
#include <libtar.h>

//dl api
#include <dlfcn.h>

//set the output functions
#ifdef DEBUG
	//use printf for 
	#define DBG_PRINT		printf
#else
	//gobble the debug up if DEBUG isn't set
	#define	DBG_PRINT(...)
#endif
//redirect to stderr
#define ERROR_PRINT(...)	fprintf(stderr, __VA_ARGS__)

//defines for license messages so i can strcmp them later
#define LICENSE_AGREE 		"I AGREE to the license."
#define LICENSE_DISAGREE 	"I DISAGREE to the license."

/**
 * where is my server?  used for downloading apps.xml and the base 
 * but not necessarily the apps, so I can download from the app provider's
 * page and apply the necessary changes in my config.
**/
#define SERVER_ROOT 		"http://download.tuxfamily.org/lportable/"

/**
 * do we use pthread or gthread?  gthread is not recommended because 
 * it blocks redraws sometimes, and it requires GNOME.
 * I'll probably withdraw farther away from GNOME in the future.
 **/
#define USE_PTHREAD			1

/** 
 * convenience macros for locking and unlocking gtk, because only
 * one thread can use gtk methods at once.
 **/
#define ENTER			gdk_threads_enter()
#define LEAVE			gdk_threads_leave()
//redraw the widget
#define REDRAW(o)		gtk_widget_queue_draw(o)

//data on a per-app basis.
typedef struct app_install_data
{
	//first 5 are obtained from apps.xml
	char* name;
	char* config;
	long long size;
	char* file;
	char* app_name;
	//these 2 are used by the program
	int chosen;
	char* error;
};

//list of apps
typedef std::vector<app_install_data*> app_list;

//adapted from http://www.linuxquestions.org/linux/articles/Technical/New_GTK_Widgets_GtkAssistant
//init data of each page.
typedef struct page_info
{
	GtkWidget* widget;
	int index;
	char* title;
	GtkAssistantPageType type;
	gboolean complete;
};

//it may say archive_data, but it's also used to hold general non-null-terminated data
typedef struct archive_data
{
	char* data;
	unsigned int len;
};

//list of data (used by the download function for each chunk, then consolidated into a single block)
typedef std::vector<archive_data> data_list;

//data for installation
typedef struct installer_label_data
{
	//the page of the selector, so I can make it complete
	GtkWidget* page;
	//the label, so I can change it accordingly
	GtkWidget* label;
	//the list of ALL apps, not just chosen
	app_list* apps;
	//the install path
	char* path;
	//the free size of the install
	long long free_space;
	//the (EXTRACTED) size of the base
	long long base_size;
	//is the install directory on a removable media?
	int mountable;
	//the release version of the base
	char* version;
};

/** callback to curl (get_data)
 * buffer - the data retreived
 * size - number of members
 * nmemb - size of a member
 * userp - NEED TO BE OF TYPE "char***" !!! 
 * 			Initially should hold the pointer to the pointer to the url.
 * 			It will be replaced with a pointer to the pointer to the copied data.
*/
static size_t installer_curl_write_callback(void* buffer, size_t size, size_t nmemb, void* userp);
/** get the (NULL TERMINATED) data of url given
 * url - where the data is
*/
static char* installer_get_data(char* url);
/** get the header of the url without getting the data
 * url - where 
*/
static char* installer_get_header(char* url);
/** get the data of the url as general dat (NON NULL TERMINATED)
 * url - where 
*/
static archive_data* installer_get_archive(char* url);
/** parse the header to find the content length
 * url - where
*/
static long long get_content_length(char* url);
/** callback to the check buttons for 
 * button - the button
 * userdata - the page the button is on, so I can set it as complete.
*/
static void installer_gtk_license_callback(GtkWidget* button, gpointer userdata);

/** make the page that displays the license
*/
static GtkWidget* make_license_page();
/** make the page where the user configures the install
 * apps - the list of apps
 * version - the install version (given in apps.xml) (put into save_data)
 * base_size - the base size of the install (given in apps.xml) (put into save_data)
 * save_data - where to put the POINTER to the data of the installation
 **/
static GtkWidget* make_selector_page(app_list* apps, char* version, long long base_size, installer_label_data** save_data);
/** make the page that displays info and the "configurers" of the installation
 * data - the data of the installation; the "order", if you will
 **/
static GtkWidget* make_installer_page(installer_label_data* data);
/** get how much free space (in bytes) is on the device with the path at dev_path
 * dev_path - where? NOT THE BLOCK FILE! (e.g. NOT "/dev/sda1", but "/" or "/media/disk") 
 **/
static long long get_free_space(char* dev_path);
/** is the path given at the root of the device it is on?
 * path - the path
 **/
static int is_at_root(char* path);
/** is the installer data valid?
 * data - installer data
 **/
static int is_valid(installer_label_data* data);
/** callback when the user presses the "Apply" button, used on the license and selector pages
 * assistant - the assistant in question
 * userdata - in this case, a pointer to the install data
 **/
static void assistant_apply_function(GtkAssistant* assistant, gpointer userdata);
/** refresh the space label (in the data given; data->label) with the data given
 * data - the install data
 **/
static void refresh_space_label(installer_label_data* data);
/** total up how much space (in bytes) is needed to install with the apps given
 * apps - the list of apps
 * base_size - the size of a basic install (no apps)
 **/
static long long get_total_size(app_list* apps, long long base_size);
/** break off another thread to install
 * data - the "order"; the install configuration
 **/
static void run_install_thread(installer_label_data* data);

/** the actual install function (it's a behemoth)
 * v_data - the install config
 **/
#ifdef USE_PTHREAD
	static void* start_install(void* v_data);
#else
	static gpointer start_install(gpointer v_data);
#endif

/** get the list of apps
 * url - where the apps are
 * version - where to save the version
*/
static app_list* get_app_list(char* url, char** version, long long* base_size);

/** callback to confirm quitting; params not used, a separate dialog is used.
*/
static void confirm_quit(GtkWidget* w, gpointer user_data);

/** dummy function to supply the pipe fd. see the definition of unpack_archive **/
static int libtar_open(const char *pathname, int oflags, ...);
/** unbzip (or ungzip) and untar the data given into the path given
 * data - the archive data retreived
 * path - where to extract
 **/
static char* unpack_archive(archive_data* data, char* path);

//the assistant, as a static so callbacks can make pages complete.
static GtkWidget* assistant;
//the agree and disagree check items from the license page, so I can disable it (MUAHAHAHHAHA NO WAY TO BACK OUT OF GPL!)
static GtkWidget* disagree;
static GtkWidget* agree;
//from the install page:
//the progress bar
static GtkWidget* progress;
//the config frame where configurers are put
static GtkWidget* config_frame;
//the "Apply" button for configurations
static GtkWidget* config_button;
//the "Redo" button in case a major (but fixable) error occured
static GtkWidget* restart_button;
//where the aforementioned buttons are put
static GtkWidget* button_hbox;

//init data for the pages
static page_info page[5] =
{
	//the intro page
	{ NULL, -1, "Introduction", GTK_ASSISTANT_PAGE_INTRO, TRUE},
	//license page
	{ NULL, -1, "License (Only applies to l-portable, not the modules)", GTK_ASSISTANT_PAGE_CONFIRM, FALSE},
	//selector page
	{ NULL, -1, "Select applications you want to install.", GTK_ASSISTANT_PAGE_CONFIRM, FALSE},
	//install page
	{ NULL, -1, "Installing...", GTK_ASSISTANT_PAGE_PROGRESS, FALSE},
	//final page
	{ NULL, -1, "Thank you for installing l-portable!", GTK_ASSISTANT_PAGE_SUMMARY, TRUE},
};

#ifndef USE_PTHREAD
	//mutex used for the following condition
	static GMutex* condition_mutex;
	//condition; used to lock up the install thread until the config button is pressed
	static GCond* config_apply_cond;
#endif

//good old main
int main(int argc, char** argv)
{
	//init thread system so we don't get weird crashes
	g_thread_init(NULL);
	gdk_threads_init();
	//unlock gdk in case something weird happened...
	LEAVE;
	//init gthread stuff now, after the threading system has been init
#ifndef USE_PTHREAD
	condition_mutex = g_mutex_new();
	config_apply_cond = g_cond_new();
#endif
	//init gtk
	gtk_init(&argc, &argv);
	//init the assistant
	assistant = gtk_assistant_new();
	//set size
	gtk_widget_set_size_request(assistant, 700, 549);
	//set title
  	gtk_window_set_title(GTK_WINDOW(assistant), "l-portable installation");
  	//get the banner on the left
  	GdkPixbuf* side = gdk_pixbuf_new_from_xpm_data(icon_side);
  	
  	//on force quit
  	g_signal_connect(G_OBJECT(assistant), "destroy", G_CALLBACK(gtk_main_quit), NULL);
  	//on "Cancel" or quit from the window frame
  	g_signal_connect(G_OBJECT(assistant), "cancel", G_CALLBACK(confirm_quit), NULL);
  	//on "Close" at the last page
  	g_signal_connect(G_OBJECT(assistant), "close", G_CALLBACK(exit), NULL);
  	
  	/** initialize pages **/
  	//init
	page[0].widget = gtk_label_new("Welcome to the installation of l-portable!\nClick Forward to continue.");
	
	//license
    page[1].widget = make_license_page();
    
    //where to save the release version
    char* version;
    //where to save the base size
    long long base_size;
    //where the apps.xml is
    char* url = new char[strlen(SERVER_ROOT) + 9];
    sprintf(url, "%sapps.xml", SERVER_ROOT);
    //get the list of apps
    app_list* list = get_app_list(url, &version, &base_size);
    //clean up
    delete url;
    
    //where to save the data
    installer_label_data* data;
    //the selector
    page[2].widget = make_selector_page(list, version, base_size, &data);
    
    //the installer
    page[3].widget = make_installer_page(data);
    
    //final summary page
    page[4].widget = gtk_label_new("Thanks for installing l-portable!");
    
    //actually add the pages
	for(int i = 0; i < 5; i++)
	{
		//add them
		page[i].index = gtk_assistant_append_page(GTK_ASSISTANT(assistant),
										page[i].widget);
		//set their titles
		gtk_assistant_set_page_title(GTK_ASSISTANT(assistant),
										page[i].widget, page[i].title);
		//set types
		gtk_assistant_set_page_type(GTK_ASSISTANT(assistant),
										page[i].widget, page[i].type);
		//are they already complete?
		gtk_assistant_set_page_complete(GTK_ASSISTANT(assistant),
										page[i].widget, page[i].complete);
  		//add the side banner for all pages
  		gtk_assistant_set_page_side_image(GTK_ASSISTANT(assistant),
  										page[i].widget, side);
	}
  	//show the assistant and every widget in it
	gtk_widget_show_all(assistant);
	//let the main loop have control (it will periodically unlock and lock)
	ENTER;
	//main loop
	gtk_main();
	//unlock if something strange happened
	LEAVE;
	//notify if main inadvertantly exited for some reason; just paranoid
	DBG_PRINT("MAIN EXITED!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
	return(0);
}

long long get_content_length(char* url)
{
	//place to put error; pcre just requires it
	static const char* error;
	static int erroffset;
	//place to put the matches
	int matches[99];
	//the regex to get the content length
	static pcre* regex = pcre_compile("^Content-Length: ([0-9]+)", 0, &error, &erroffset, NULL);
	//get the header
	char* head = installer_get_header(url);
	
	//the number of times the regex matches...
	int result = pcre_exec(regex, NULL, head, strlen(head)
				, 0, 0, matches, 99);
	//over 2 (one for "Content-Length: ", one for the actual length) (YES!)
	if(result >= 2)
	{
		//save the result string into a temp..
		char* tmp = new char[matches[3] - matches[2] + 1];
		sprintf(tmp, "%.*s", (matches[3] - matches[2]), (head + matches[2]));
		//and resolve it into a long long, just because there might be a large download
		long long ret = atoll(tmp);
		//debug print
		DBG_PRINT("MATCH: %s, %i\n", tmp, ret);
		//clean up and return
		delete[] tmp;
		return(ret);
	}
	//under 2 (NO)
	DBG_PRINT("NO MATCH???: %i\n", result);
	//complain
	return(-1);
}

app_list* get_app_list(char* url, char** version, long long* base_size)
{
	//the document
	TiXmlDocument* doc = new TiXmlDocument();
	//download from the url and parse from that
	if(doc->Parse(installer_get_data(url)) != NULL)
	{
		TiXmlElement* root = doc->RootElement();
		if(root != NULL)
		{
			//what format version is the list file?
			int list_version;
			//try top get the version...
			if(root->QueryIntAttribute("meta_version", &list_version) == TIXML_SUCCESS)
			{
				//yes, debug print
				DBG_PRINT("\tlist format version: %d\n", list_version);
				const char* tmp = root->Attribute("release_version");
				if(tmp != NULL)
				{
					*version = new char[strlen(tmp) + 1];
					strcpy(*version, tmp);
					
					*base_size = atoll(root->Attribute("extract_size"));
					switch(list_version)
					{
						case 1:
						{
							//count the number of apps
							app_list* list = new app_list();
							TiXmlNode* first = root->FirstChild();
							if(first)
							{
								TiXmlNode* current = first;
								while(current != NULL)
								{
									if(current->Type() == TiXmlNode::ELEMENT)
									{
										TiXmlElement* e = current->ToElement();
										if(!strcmp(e->Value(), "app"))
										{
											app_install_data* data = new(app_install_data);
											tmp = e->GetText();
											if(tmp != NULL)
											{
												data->name = new char[strlen(tmp) + 1];
												strcpy(data->name, tmp);
												DBG_PRINT("\tname: %s\n", data->name);
												
												tmp = e->Attribute("app_name");
												if(tmp != NULL)
												{
													data->app_name = new char[strlen(tmp) + 1];
													strcpy(data->app_name, tmp);
													DBG_PRINT("\tapp_name: %s\n", data->app_name);
													tmp = e->Attribute("file");
													if(tmp != NULL)
													{
														data->file = new char[strlen(tmp) + 1];
														strcpy(data->file, tmp);
														DBG_PRINT("\tfile: %s\n", data->file);
														tmp = e->Attribute("extract_size");
														if(tmp != NULL)
														{
															data->size = atoll(tmp);
															DBG_PRINT("\tsize: %lld\n", data->size);
															tmp = e->Attribute("config");
															if(tmp != NULL)
															{
																data->config = new char[strlen(tmp) + 1];
																strcpy(data->config, tmp);
																DBG_PRINT("\tconfig: %s\n", data->config);
															}
															else
																data->config = NULL;
															
															data->chosen = 0;
															data->error = NULL;
															list->push_back(data);
														}
														/** if i can find the extract size **/
														else
															delete data;
													}
													/** if i can find the file **/
													else
														delete data;
												}
												/** if i can find the app_name **/
												else
													delete data;
											}
											/** if i can get the name **/
											else
												delete data;
										}
										/** if its tag is "app" **/
									}
									/** if the node is a element **/
									//go to the next one
									current = current->NextSibling();
								}
								/** for each app **/
							}
							/** if i can get the first sibling of the apps **/
							//clean up and return the list (even if it is empty)
							delete doc;
							return(list);
							break;
						}
						/** the list version is 1 **/
					}
					/** switch(list_version) **/
				}
				/** if i can find the release version **/
			}
			/** if i can get the list version **/
		}
		/** if i can get the root **/
	}
	/** if the apps.xml is valid **/
	//can't parse?
	if(doc)	
		delete doc;
	return(NULL);
}

void assistant_apply_function(GtkAssistant* assistant, gpointer userdata)
{
	DBG_PRINT("\"apply\" button pressed: page: %i\n", gtk_assistant_get_current_page(assistant));
	int page = gtk_assistant_get_current_page(assistant);
	if(page == 2)
	{
		gtk_widget_set_sensitive(disagree, false);
		gtk_widget_set_sensitive(agree, false);
	}
	else if(page == 3)
	{
		installer_label_data* label = (installer_label_data*)userdata;
		if(!is_valid(label))
		{
			/* create the dialog */
			GtkWidget* dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
							GTK_MESSAGE_ERROR, GTK_BUTTONS_YES_NO,
							"The install path either does not exist or is not a directory...");

			gtk_window_set_title(GTK_WINDOW(dialog), "Invalid Install Configuration");

			/* run the dialog */
			gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);
		}
		else
		{
			run_install_thread(label);
		}
	}
}

static std::string buffer;

size_t installer_curl_write_callback(void* data, size_t size, size_t nmemb, void* userp)
{
	if(data != NULL)
	{
		buffer.append((char*)data, size * nmemb);
		return(size * nmemb);
	}
	return(0);
}

static data_list buffer_list;

size_t installer_curl_archive_write_callback(void* data, size_t size, size_t nmemb, void* userp)
{
	//debug print
	//DBG_PRINT("Transfer size: %i * %i = %i\n", size, nmemb, size * nmemb);
	//initialize the copy, save room for null char
	char* ret = new char[size * nmemb];
	//copy it over
	memcpy(ret, data, size * nmemb);
	archive_data add = {ret, (unsigned int)(size * nmemb)};
	buffer_list.push_back(add);
	return(size * nmemb);
}

char* installer_get_data(char* url)
{
	DBG_PRINT("URL: %s\n", url);
	
	static char error_buf[CURL_ERROR_SIZE];
	CURL* handle = curl_easy_init();
	
	curl_easy_setopt(handle, CURLOPT_URL, url);
	curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, installer_curl_write_callback);
	curl_easy_setopt(handle, CURLOPT_WRITEDATA, NULL);
	curl_easy_setopt(handle, CURLOPT_ERRORBUFFER, error_buf);
	int success = curl_easy_perform(handle);
	curl_easy_cleanup(handle);
	if(success == 0)
	{
		DBG_PRINT("Download Successful\n");
		char* ret = new char[buffer.size()];
		strcpy(ret, buffer.c_str());
		buffer.clear();
		return(ret);
	}
	else
	{
		DBG_PRINT("Download Error: %s\n", error_buf);
		return(NULL);
	}
}

archive_data* installer_get_archive(char* url)
{
	DBG_PRINT("URL: %s\n", url);
	
	static char error_buf[CURL_ERROR_SIZE];
	CURL* handle = curl_easy_init();
	
	curl_easy_setopt(handle, CURLOPT_URL, url);
	curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, installer_curl_archive_write_callback);
	curl_easy_setopt(handle, CURLOPT_WRITEDATA, NULL);
	curl_easy_setopt(handle, CURLOPT_ERRORBUFFER, error_buf);
	int success = curl_easy_perform(handle);
	curl_easy_cleanup(handle);
	if(success == 0)
	{
		DBG_PRINT("Download Successful\n");
		data_list::iterator it = buffer_list.begin();
		unsigned int length = 0;
		while(it != buffer_list.end())
		{
			length += it->len;
			it++;
		}
		archive_data* ret = new(archive_data);
		ret->len = length;
		ret->data = new char[length];
		unsigned int pos = 0;
		it = buffer_list.begin();
		while(it != buffer_list.end())
		{
			length = it->len;
			memcpy(ret->data + pos, it->data, length);
			pos += length;
			it++;
		}
		buffer_list.clear();
		return(ret);
	}
	else
	{
		DBG_PRINT("Download Error: %s\n", error_buf);
		return(NULL);
	}
}

char* installer_get_header(char* url)
{
	DBG_PRINT("URL: %s\n", url);
	
	static char error_buf[CURL_ERROR_SIZE];
	CURL* handle = curl_easy_init();
	
	curl_easy_setopt(handle, CURLOPT_URL, url);
	curl_easy_setopt(handle, CURLOPT_HEADER, true);
	curl_easy_setopt(handle, CURLOPT_NOBODY, true);
	curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, installer_curl_write_callback);
	curl_easy_setopt(handle, CURLOPT_WRITEDATA, NULL);
	curl_easy_setopt(handle, CURLOPT_ERRORBUFFER, error_buf);
	int success = curl_easy_perform(handle);
	curl_easy_cleanup(handle);
	if(success == 0)
	{
		DBG_PRINT("Download Successful\n");
		char* ret = new char[buffer.size()];
		strcpy(ret, buffer.c_str());
		buffer.clear();
		DBG_PRINT("Data: %s\n", ret);
		return(ret);
	}
	else
	{
		DBG_PRINT("Download Error: %s\n", error_buf);
		return(NULL);
	}
}

#define	PIPE_READ 				0
#define PIPE_WRITE				1
#define SUCCESS_EXTRACTING		"Success extracting"
static int parent_write[2];
static int child_write[2];
int libtar_open(const char *pathname, int oflags, ...)
{
	DBG_PRINT("libtar open\n");
	return(parent_write[PIPE_READ]);
}

static char* unpack_archive(archive_data* data, char* path)
{
	//guess a length
	unsigned int tar_archive_size = data->len * 10;
	//initialize
	char* tar_archive = new char[tar_archive_size];
	//first try to decompress
	int res = BZ2_bzBuffToBuffDecompress(tar_archive, &tar_archive_size, data->data, data->len, 0, 4);
	
	//if it is too large, try again with a large size
	while(res == BZ_OUTBUFF_FULL)
	{
		delete[] tar_archive;
		tar_archive_size += data->len;
		tar_archive = new char[tar_archive_size];
		res = BZ2_bzBuffToBuffDecompress(tar_archive, &tar_archive_size, data->data, data->len, 0, 4);
	}
	//remove the old data, regardless of success
	delete data;
	//if there was some other error than overflow
	if(res != BZ_OK)
	{
		DBG_PRINT("Error extracting from bzip file: %s\n", strerror(errno));
		//delete and quit
		delete[] tar_archive;
		return(strerror(errno));
	}
	//otherwise, let us know
	DBG_PRINT("\nSuccess extracting from bzip\n");
	
	//try to open a pipe, because we're going need it...
	if(pipe(parent_write) < 0 || pipe(child_write) < 0)
	{
		DBG_PRINT("Error opening pipe: %s\n", strerror(errno));
		//if not successful, delete and quit
		delete[] tar_archive;
		return(strerror(errno));
	}
	//otherwise, let us know
	DBG_PRINT("Success opening pipe\n");
	
	//libtar_open will give tar_open the pipe read file descriptor no matter what...
	tartype_t tar_funcs = 
	{
		libtar_open,
		close,
		read,
		write
	};
	
	//initialize
	TAR* t;
	//give some random values; only the first and third parameters matter
	if(tar_open(&t, "/", &tar_funcs, O_WRONLY | O_CREAT, 0644, 4) == -1)
	{
		DBG_PRINT("Error with libtar: %s\n", strerror(errno));
		//if not successful, delete and quit
		delete[] tar_archive;
		return(strerror(errno));
	}
	//otherwise, let us know
	DBG_PRINT("Success opening tar\n");
	
	//pid for forking; need to know who's child
	int pid;
	if((pid = fork()) < 0)
	{
		DBG_PRINT("Error forking: %s\n", strerror(errno));
		//if not successful, delete and quit
		delete[] tar_archive;
		return(strerror(errno));
	}
	else
	{
		//if I'm the child...
		if(pid == 0)
		{
			close(child_write[PIPE_READ]);
			close(parent_write[PIPE_WRITE]);
			
			//extract it
			//if there is an error, I can't know, though...
			if(tar_extract_all(t, path) == -1)
			{
				char* error = new char[23 + strlen(strerror(errno))];
				sprintf(error, "Problem extracting: %s", strerror(errno));
				write(child_write[PIPE_WRITE], error, (strlen(error) + 1) * sizeof(char));
			}
			else
			{
				const char* response = SUCCESS_EXTRACTING;
				write(child_write[PIPE_WRITE], response, (strlen(response) + 1) * sizeof(char));
			}
			//close the write
			close(child_write[PIPE_WRITE]);
			
			//close (will close the read)
			tar_close(t);
			
			//exit to prevent two processes from doint the same thing...
			_exit(0);
			DBG_PRINT("Hopefully I won't see this...\n");
		}
		else
		{	
			
			//write the archive in
			write(parent_write[PIPE_WRITE], tar_archive, tar_archive_size);
			//let us know when you finished...
			DBG_PRINT("Finished writing to pipe\n");
			//close the write
			close(parent_write[PIPE_WRITE]);
			close(child_write[PIPE_WRITE]);
			close(parent_write[PIPE_READ]);
			
			char* buffer = new char[1024];
			int size = read(child_write[PIPE_READ], buffer, 1024);
			DBG_PRINT("Response: %s\n", buffer);
			//close the read
			close(child_write[PIPE_READ]);
			
			//wait for child to finish reading
			waitpid(pid, NULL, 0);
			
			if(strcmp(buffer, SUCCESS_EXTRACTING))
			{
				delete[] tar_archive;
				return(buffer);
			}
			delete[] buffer;
		}
	}
	DBG_PRINT("Success extracting archive to %s, returning\n", path);
	//delete and return as a hero
	delete[] tar_archive;
	return(NULL);
}

//adapted from http://www.gtkforums.com/about672.html
void confirm_quit(GtkWidget *w, gpointer user_data)
{
	/* create the dialog */
	GtkWidget* dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
					"Are you sure you want to quit?");

	gtk_window_set_title(GTK_WINDOW(dialog), "Really Quit?");

	/* run the dialog */
	int response = gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);   
	if(response == GTK_RESPONSE_YES)
	{
		exit(0);
	}
}

void installer_gtk_license_callback(GtkWidget* button, gpointer userdata)
{
	DBG_PRINT("label: %s\n", gtk_button_get_label(GTK_BUTTON(button)));
	if(!strcmp(gtk_button_get_label(GTK_BUTTON(button)), LICENSE_AGREE))
		gtk_assistant_set_page_complete(GTK_ASSISTANT(assistant), (GtkWidget*)userdata, true);
	else
		gtk_assistant_set_page_complete(GTK_ASSISTANT(assistant), (GtkWidget*)userdata, false);
}

GtkWidget* make_license_page()
{
	GtkWidget* ret = gtk_vbox_new(FALSE, 0);
	GtkWidget* scroll = gtk_scrolled_window_new(NULL, NULL);
    GtkTextBuffer* buffer = gtk_text_buffer_new(NULL);
    gtk_text_buffer_set_text(GTK_TEXT_BUFFER(buffer), gpl, gpl_len);
    GtkWidget* text_view = gtk_text_view_new_with_buffer(buffer);
    gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view), false);
    gtk_container_add(GTK_CONTAINER(scroll), text_view);
    gtk_widget_set_size_request(scroll, -1, 300);
    gtk_container_add(GTK_CONTAINER(ret), scroll);
    /** agree (hopefully) **/
    GtkWidget* vbox2 = gtk_vbox_new(TRUE, 0);
    disagree = gtk_radio_button_new_with_label(NULL, LICENSE_DISAGREE);
    gtk_widget_set_size_request(disagree, -1, 10);
    gtk_container_add(GTK_CONTAINER(vbox2), disagree);
    agree = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(disagree)),
    					LICENSE_AGREE);
    gtk_widget_set_size_request(agree, -1, 10);
    gtk_container_add(GTK_CONTAINER(vbox2), agree);
    gtk_widget_set_size_request(vbox2, -1, 22);
    gtk_container_add(GTK_CONTAINER(ret), vbox2);
    
    g_signal_connect(G_OBJECT(disagree), "clicked", 
    		G_CALLBACK(installer_gtk_license_callback), (gpointer)ret);
    g_signal_connect(G_OBJECT(agree), "clicked", 
    		G_CALLBACK(installer_gtk_license_callback), (gpointer)ret);
    return(ret);
}

void installer_gtk_file_chooser_callback(GtkFileChooser* button, gpointer userdata)
{
	installer_label_data* label_data = (installer_label_data*)userdata;
	label_data->path = gtk_file_chooser_get_filename(button);
	DBG_PRINT("path: %s\n", label_data->path);
	label_data->free_space = get_free_space(label_data->path);
	label_data->mountable = is_at_root(label_data->path);
	refresh_space_label(label_data);
}

void installer_gtk_item_check_callback(GtkToggleButton* button, gpointer userdata)
{
	installer_label_data* label_data = (installer_label_data*)userdata;
	char* label = const_cast<char*>(gtk_button_get_label(GTK_BUTTON(button)));
	DBG_PRINT("Looking for %s\n", label);
	app_list::iterator it = label_data->apps->begin();
	while(it != label_data->apps->end())
	{
		if(!strcmp((*it)->name, label))
		{
			DBG_PRINT("Found %s, setting...\n", (*it)->name);
			(*it)->chosen = gtk_toggle_button_get_active(button);
		}
		it++;
	}
	refresh_space_label(label_data);
}

GtkWidget* make_selector_page(app_list* apps, char* version, long long base_size, installer_label_data** save_data)
{
	//DBG_PRINT("version: %s\n", version);
	
	GtkWidget* ret = gtk_vbox_new(FALSE, 0);
	GtkWidget* space_label = gtk_label_new("");
	gtk_widget_set_size_request(space_label, -1, 20);
	installer_label_data* label_data = new(installer_label_data);
	label_data->page = ret;
	label_data->label = space_label;
	label_data->apps = apps;
	label_data->path = "/home/jeff/workspace/l-portable/test";
	label_data->free_space = get_free_space(label_data->path);
	label_data->mountable = is_at_root(label_data->path);
	label_data->version = version;
	label_data->base_size = base_size;
	refresh_space_label(label_data);
	
	*save_data = label_data;
	
	g_signal_connect(G_OBJECT(assistant), "apply", G_CALLBACK(assistant_apply_function), (gpointer)(label_data));
	
	GtkWidget* file_selector_frame = gtk_frame_new("Select the install directory.");
	gtk_frame_set_shadow_type(GTK_FRAME(file_selector_frame), GTK_SHADOW_IN);
	GtkWidget* file_selector = gtk_file_chooser_button_new("Select the install directory."
							, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(file_selector), "/home/jeff/workspace/l-portable/test");
	gtk_widget_set_size_request(file_selector, -1, 10);
	g_signal_connect(G_OBJECT(file_selector), "current-folder-changed", 
    		G_CALLBACK(installer_gtk_file_chooser_callback), (gpointer)(label_data));
    
	gtk_container_add(GTK_CONTAINER(file_selector_frame), file_selector);
	gtk_widget_set_size_request(file_selector_frame, -1, 10);
	gtk_container_add(GTK_CONTAINER(ret), file_selector_frame);
	
	gtk_container_add(GTK_CONTAINER(ret), space_label);
	
	GtkWidget* app_selector_frame = gtk_frame_new("Select the applications to install.");
	gtk_frame_set_shadow_type(GTK_FRAME(app_selector_frame), GTK_SHADOW_IN);
	GtkWidget* list = gtk_scrolled_window_new(NULL, NULL);
	GtkWidget* box = gtk_vbox_new(FALSE, 0);
	GtkWidget* each_item;
	app_list::iterator it = apps->begin();
	while(it != apps->end())
	{
		each_item = gtk_check_button_new_with_label((*it)->name);
		g_signal_connect(G_OBJECT(each_item), "toggled", 
    		G_CALLBACK(installer_gtk_item_check_callback), (gpointer)(label_data));
		gtk_container_add(GTK_CONTAINER(box), each_item);
		it++;
	}
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(list), box);
	gtk_container_add(GTK_CONTAINER(app_selector_frame), list);
	gtk_container_add(GTK_CONTAINER(ret), app_selector_frame);
	return(ret);
}

//is this a directory?
int is_dir(const char* fname)
{
	//using open and fstat
	int fd = open(fname, O_RDONLY);
	struct stat *file_stat = new(struct stat);
	int exists = !fstat(fd, file_stat);	//exists?
	close(fd);
	if(exists)
	{
		int ret = S_ISDIR(file_stat->st_mode);
		delete file_stat;
		return(ret);
	}
	delete file_stat;
	return(0);
}

int is_valid(installer_label_data* data)
{
	return(data->path != NULL && is_dir(data->path) && (data->free_space >= get_total_size(data->apps, data->base_size)));
}

long long get_total_size(app_list* apps, long long base_size)
{
	long long ret = base_size;
	app_list::iterator it = apps->begin();
	while(it != apps->end())
	{
		DBG_PRINT("%s: %lld, %i\n", (*it)->name, (*it)->size, (*it)->chosen);
		if((*it)->chosen)
		{
			
			ret += (*it)->size;
		}
		it++;
	}
	return(ret);
}

void refresh_space_label(installer_label_data* data)
{
	DBG_PRINT("data: %p, data->apps: %p\n", data, data->apps);
	long long total_size = get_total_size(data->apps, data->base_size);
	char message[1024];	//arbitrary number, must be more exact?
	/** messages:
	 * This installation will require xxxMB out of xxxMB[.|, NOT ENOUGH SPACE!]
	 * [The volume you have chosen is at the root of a removable media;
	 * you cannot use special functions if you install there.]
	**/
	if(data->path == NULL)
	{
		sprintf(message, "%s %.3Lf %s", "This installation will require", (long double)total_size / 1024.0 / 1024.0
			, "MB; Please choose a install directory.");
	}
	else
	{
		sprintf(message, "%s %.3Lf %s %.3Lf %s%s"
			, "This installation will require"
			, (long double)total_size / 1024.0 / 1024.0
			, "MB out of"
			, (long double)data->free_space / 1024.0 / 1024.0
			, total_size > data->free_space ? "MB, NOT ENOUGH SPACE!" : "MB."
			, data->mountable ? "" : "\nThe volume you have chosen is not at the root of a removable media;\nyou cannot use special functions if you install there.");
	}
	gtk_label_set_text(GTK_LABEL(data->label), message);
	gtk_assistant_set_page_complete(GTK_ASSISTANT(assistant), data->page, is_valid(data));
}

int is_at_root(char* path)
{
	GFile* f = g_file_new_for_path(path);
	GMount* enc = g_file_find_enclosing_mount(f, NULL, NULL);
	//if there is an enclosing volume and the path given is the root
	return(enc != NULL && !strcmp(path, g_file_get_path(g_mount_get_root(enc))));
}

long long get_free_space(char* dev_path)
{
	struct statfs sfs;
	if(statfs(dev_path, &sfs) >= 0)
	{
		DBG_PRINT("sfs bsize: %i, sfs bfree: %i\n", sfs.f_bsize, sfs.f_bavail);
		long long int ret= (long long)sfs.f_bsize * (long)sfs.f_bavail;
		DBG_PRINT("Free space of %s: %lld (%f Gb)\n", dev_path, ret, (ret / 1024.0 / 1024 / 1024));
		return(ret);
	}
	DBG_PRINT("strerror: %s\n", strerror(errno));
	return(-1);
}

void run_install_thread(installer_label_data* data)
{
#ifdef USE_PTHREAD
	pthread_t thread;
	int response = pthread_create(&thread, NULL, start_install, (void*)data);
#else
	GThread* thread = g_thread_create(start_install, (void*)data, true, NULL);
#endif
}

#ifdef USE_PTHREAD
	static pthread_mutex_t condition_mutex 		= PTHREAD_MUTEX_INITIALIZER;
	static pthread_cond_t  config_apply_cond	= PTHREAD_COND_INITIALIZER;
	void* start_install(void* v_data)
#else
	gpointer start_install(gpointer v_data)
#endif
{
	installer_label_data* data = (installer_label_data*)v_data;
	
	ENTER;
	gtk_widget_set_sensitive(restart_button, false);
	REDRAW(restart_button);
	gtk_widget_set_sensitive(config_button, false);
	REDRAW(config_button);
	LEAVE;
	
	DBG_PRINT("starting install...\n");
	//number of steps to take; 2 is for downloading then installing the base
	int num_steps = 2;
	//number of steps acutally taken
	int steps = 0;
	int max_folder_name_size = 0;
	app_list::iterator it = data->apps->begin();
	while(it != data->apps->end())
	{
		if((*it)->chosen)
		{
			if(strlen((*it)->app_name) > max_folder_name_size)
				max_folder_name_size = strlen((*it)->app_name);
			//downloading then extracting the module
			num_steps += 3;
			/** downloading, having the user do the config
			 * then actually performing the config
			 * (might take a long time for mozilla profiles)
			**/
			if((*it)->config)
				num_steps += 3;
			}
		it++;
	}
	DBG_PRINT("number of install steps: %i\n", num_steps);
	
	//do the actual install!
	
	//(1) download the base...
	//figure out what the base file is first...
	ENTER;
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), "Downloading the base... (0%)");
	REDRAW(progress);
	LEAVE;
	
	char* url = new char[strlen(SERVER_ROOT) + strlen(data->version) + 26];
    sprintf(url, "%sl-portable_base-%s.tar.bz2", SERVER_ROOT, data->version);
    DBG_PRINT("base url: %s\n", url);
    archive_data* base = installer_get_archive(url);
    delete url;
    if(base == NULL)
    {
    	ENTER;
    	gtk_assistant_set_page_type(GTK_ASSISTANT(assistant), page[3].widget, GTK_ASSISTANT_PAGE_CONTENT);
    	REDRAW(page[3].widget);
    	gtk_widget_set_sensitive(restart_button, true);
    	REDRAW(restart_button);
    	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), "Error downloading the base...Check your\ninternet connection, and resdo the install");
    	REDRAW(progress);
    	LEAVE;

    	return(NULL);
	}
	else
		steps++;
    
    //(2) extract the base...
    char* progress_str = new char[1024];
    
    sprintf(progress_str, "Extracting the base... (%.3f%%)", (double)steps / num_steps * 100);
    
    ENTER;
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), progress_str);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), (double)steps / num_steps);
    REDRAW(progress);
    LEAVE;
    
    delete progress_str;
    char* extract_response = unpack_archive(base, data->path);
    //delete base;
    if(extract_response != NULL)
    {
    	ENTER;
    	gtk_assistant_set_page_type(GTK_ASSISTANT(assistant), page[3].widget, GTK_ASSISTANT_PAGE_CONTENT);
    	REDRAW(page[3].widget);
    	gtk_widget_set_sensitive(restart_button, true);
    	REDRAW(restart_button);
    	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), "Error extracting the base...Check your install directory\nfor permissions, and redo the install");
    	REDRAW(progress);
    	LEAVE;
    	
    	return(NULL);	
	}
	else
		steps++;
	
	//(3) for each app...
	char* app_folder = new char[strlen(data->path) + max_folder_name_size + 13];
	it = data->apps->begin();
	while(it != data->apps->end())
	{
		if((*it)->chosen)
		{
			(*it)->error = "Program \"%s\" did not have any errors in installation.\n";
			
			//(1) Download archive
			DBG_PRINT("program: %s\n", (*it)->name);
			progress_str = new char[strlen((*it)->name) + 200];
			sprintf(progress_str, "Download program \"%s\"... (%.3f%%)", (*it)->name, (double)steps / num_steps * 100);
			
			ENTER;
			gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), progress_str);
			gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), (double)steps / num_steps);
			REDRAW(progress);
			LEAVE;
			
			archive_data* archive = installer_get_archive((*it)->file);
			if(archive != NULL)
			{
				//(2) Make install directory
				steps++;
				sprintf(app_folder, "%s/l-portable/%s", data->path, (*it)->app_name);
			
				delete progress_str;
				progress_str = new char[strlen(app_folder) + 200];
				sprintf(progress_str, "Making folder \"%s\"... (%.3f%%)", app_folder, (double)steps / num_steps * 100);
				
				ENTER;
				gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), progress_str);
				gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), (double)steps / num_steps);
				REDRAW(progress);
				LEAVE;
				
				DBG_PRINT("\napp_folder: %s\n", app_folder);
				if(mkdir(app_folder, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == 0)
				{
					//(3) Extract files
					steps++;
					
					delete progress_str;
					progress_str = new char[strlen((*it)->name) + 200];
					sprintf(progress_str, "Installing program \"%s\"... (%.3f%%)", (*it)->name, (double)steps / num_steps * 100);
					
					ENTER;
					gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), progress_str);
					gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), (double)steps / num_steps);
					REDRAW(progress);
					LEAVE;
					
					extract_response = unpack_archive(archive, app_folder);
					//delete archive;
					if(extract_response == NULL)
					{
						steps++;
						if((*it)->config)
						{
							//(4) Download configurer
							sprintf(progress_str, "Downloading configurer program \"%s\"... (%.3f%%)", (*it)->name, (double)steps / num_steps * 100);
							
							ENTER;
							gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), progress_str);
							gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), (double)steps / num_steps);
							REDRAW(progress);
							LEAVE;
							
							//even though it says get archive, it really means get all sorts of info...
							//disregard the other imposter...
							archive_data* config = installer_get_archive((*it)->config);
							if(config != NULL)
							{
								char* config_path = tmpnam(NULL);
								DBG_PRINT("config path: %s\n", config_path);
								int config_fd = creat(config_path, S_IRWXU);
								if(config_fd != -1)
								{
									if(write(config_fd, config->data, config->len) != -1)
									{
										close(config_fd);
										void* so_handle = dlopen(config_path, RTLD_NOW);
										if(so_handle != NULL)
										{
											typedef char* (*widget_maker)(GtkWidget**);
											char* (*make_config_widget)(GtkWidget**) = (widget_maker)dlsym(so_handle, "make_config_widget");
											if(make_config_widget != NULL)
											{
												GtkWidget* configurer = NULL;
												char* response = make_config_widget(&configurer);
												if(response == NULL)
												{
													ENTER;
													gtk_container_add(GTK_CONTAINER(config_frame), configurer);
													gtk_widget_show_all(configurer);
													gtk_widget_set_sensitive(config_frame, true);
													gtk_widget_set_sensitive(config_button, true);
													//REDRAW(config_button);
													//REDRAW(config_frame);
													//REDRAW(configurer);
													//REDRAW(assistant);
													LEAVE;
													
													//(5) User input
													steps++;
													sprintf(progress_str, "Configure program \"%s\" below... (%.3f%%)", (*it)->name, (double)steps / num_steps * 100);
													
													ENTER;
													gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), (double)steps / num_steps);
													gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), progress_str);
													REDRAW(progress);
													LEAVE;
													
#ifdef USE_PTHREAD
													//wait...for config button to be pressed
													pthread_mutex_lock(&condition_mutex);
													pthread_cond_wait(&config_apply_cond, &condition_mutex);
													pthread_mutex_unlock(&condition_mutex);
#else
													g_mutex_lock(condition_mutex);
													g_cond_wait(config_apply_cond, condition_mutex);
													g_mutex_unlock(condition_mutex);
#endif
													
													
													//(6) Perform the config
													steps++;
													sprintf(progress_str, "Configuring the program \"%s\"... (%.3f%%)", (*it)->name, (double)steps / num_steps * 100);
													
													ENTER;
													gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), (double)steps / num_steps);
													gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), progress_str);
													REDRAW(progress);
													LEAVE;
													
													typedef char* (*config_performer)(char*);
													char* (*do_config)(char*) = (config_performer)dlsym(so_handle, "do_config");
													if(do_config != NULL)
													{
														response = do_config(app_folder);
														steps++;
														if(response != NULL)
														{
															DBG_PRINT("\terror in do_config: %s\n", response);
															(*it)->error = new char[strlen(response) + 55];
															sprintf((*it)->error, "Error performing configuration for program \"%%s\": %s\n", response);
														}
														//error in do_config;
													}
													else
													{
														DBG_PRINT("\terror getting performation func: %s\n", dlerror());
														(*it)->error = "Error performing configuration for program \"%s\"\n";
														steps++;
													}
													//can't get config performer
													
													ENTER;
													gtk_container_remove(GTK_CONTAINER(config_frame), configurer);
													gtk_widget_set_sensitive(config_frame, false);
													gtk_widget_set_sensitive(config_button, false);
													REDRAW(config_frame);
													REDRAW(config_button);
													LEAVE;
												}
												else
												{
													DBG_PRINT("\terror in init func: %s\n", response);
													(*it)->error = new char[strlen(response) + 55];
													sprintf((*it)->error, "Error initializating configurer for program \"%%s\": %s\n", response);
													steps += 3;
												}
												//error in init?
											}
											else
											{
												DBG_PRINT("\terror getting widget init func: %s\n", dlerror());
												(*it)->error = "Error initializating configurer for program \"%s\"\n";
												steps += 3;
											}
											//couldn't get widget initialization function
											dlclose(so_handle);
										}
										else
										{
											DBG_PRINT("\terror opening config: %s\n", dlerror());
											(*it)->error = "Error opening configurer for program \"%s\"\n";
											steps += 3;
										}
										//couldn't open config
										remove(config_path);
									}
									else
									{
										close(config_fd);
										remove(config_path);
										DBG_PRINT("\terror writing to config: %s\n", strerror(errno));
										(*it)->error = "Error writing to temporary file for program \"%s\"\n";
										steps += 3;
									}
									//couldn't write
								}
								else
								{
									DBG_PRINT("\terror creating config: %s\n", strerror(errno));
									(*it)->error = "Error creating temporary file for program \"%s\"\n";
									steps += 3;
								}
								//could create temp file
							}
							else
							{
								DBG_PRINT("\terror downloading config\n");
								(*it)->error = "Error downloading configurer for program \"%s\"\n";
								steps += 3;
							}
							//downloading config didn't succeed
						}
					}
					else
					{
						DBG_PRINT("\terror extracting\n");
						(*it)->error = "Error extracting program \"%s\"\n";
						steps += 1;
						if((*it)->config)
						{
							steps += 3;
						}
					}
					//extract did not succeed
				}
				else
				{
					DBG_PRINT("\terror making folder\n");
					(*it)->error = "Error making folder\"%s\"\n";
					steps += 1;
					if((*it)->config)
					{
						steps += 3;
					}
				}
				//mkdir didn't succeed
				
			} /** if(archive != NULL) **/
			else
			{
				DBG_PRINT("\terror downloading\n");
				(*it)->error = "Error downloading program \"%s\"\n";
				steps += 2;
				if((*it)->config)
				{
					steps += 3;
				}
			}
			//downloading did not succeed
			delete progress_str;
		} /** if((*it)->chosen) **/
		it++;
	} /** while(it != data->apps->end()) **/
	
	ENTER;
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progress), 1.00);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progress), "Finished installation!"); 
	gtk_assistant_set_page_complete(GTK_ASSISTANT(assistant), page[3].widget, true);
	
    GtkTextBuffer* buffer = gtk_text_buffer_new(NULL);
    LEAVE;
    
	it = data->apps->begin();
	while(it != data->apps->end())
	{
		if((*it)->error != NULL)
		{
			char* error = new char[strlen((*it)->error) + strlen((*it)->name) + 1];
			sprintf(error, (*it)->error, (*it)->name);
			ENTER;
			gtk_text_buffer_insert_at_cursor(buffer, error, -1);
			LEAVE;
			delete[] error;
		}
		it++;
	}
	ENTER;
	GtkTextIter start, end;
	gtk_text_buffer_get_start_iter(buffer, &start);
	gtk_text_buffer_get_end_iter(buffer, &end);
	DBG_PRINT("\n<errors>\n%s</errors>\n", gtk_text_buffer_get_text(buffer,
					&start, &end, true));
	
	GtkWidget* scroll = gtk_scrolled_window_new(NULL, NULL);
    GtkWidget* text_view = gtk_text_view_new_with_buffer(buffer);
    gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view), false);
    gtk_widget_show_all(text_view);
    gtk_widget_show_all(scroll);
    gtk_widget_set_sensitive(config_frame, true);
    
    gtk_container_add(GTK_CONTAINER(scroll), text_view);
    gtk_frame_set_label(GTK_FRAME(config_frame), "Errors");
    gtk_container_add(GTK_CONTAINER(config_frame), scroll);
	LEAVE;

	delete app_folder;
	return(NULL);
}

void restart_install(GtkButton* button, gpointer userdata)
{
	run_install_thread((installer_label_data*)userdata);
}

void config_applied(GtkButton* b, gpointer userdata)
{
#ifdef USE_PTHREAD
	pthread_cond_broadcast(&config_apply_cond);
#else
	g_cond_broadcast(config_apply_cond);
#endif
}

GtkWidget* make_installer_page(installer_label_data* data)
{
	GtkWidget* ret = gtk_vbox_new(FALSE, 2);
	
	progress = gtk_progress_bar_new();
	gtk_widget_set_size_request(progress, -1, 80);
	gtk_container_add(GTK_CONTAINER(ret), progress);
	
	config_frame = gtk_frame_new("Configuration");
	gtk_frame_set_shadow_type(GTK_FRAME(config_frame), GTK_SHADOW_IN);
	gtk_widget_set_sensitive(config_frame, false);
	gtk_widget_set_size_request(config_frame, -1, 380);
	/*
	config_vbox = gtk_vbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(config_vbox), gtk_label_new("LOL!!!"));
	gtk_container_add(GTK_CONTAINER(config_frame), config_vbox);
	// */
	gtk_container_add(GTK_CONTAINER(ret), config_frame);
		
	button_hbox = gtk_hbox_new(FALSE, 2);
	
	restart_button = gtk_button_new_from_stock(GTK_STOCK_REDO);
	gtk_widget_set_sensitive(restart_button, false);
	gtk_widget_set_size_request(restart_button, -1, 50);
	gtk_container_add(GTK_CONTAINER(button_hbox), restart_button);
	g_signal_connect(G_OBJECT(restart_button), "clicked", G_CALLBACK(restart_install), (gpointer)data);
	
	config_button = gtk_button_new_from_stock(GTK_STOCK_APPLY);
	gtk_widget_set_sensitive(config_button, false);
	gtk_widget_set_size_request(config_button, -1, 50);
	gtk_container_add(GTK_CONTAINER(button_hbox), config_button);
	g_signal_connect(G_OBJECT(config_button), "clicked", G_CALLBACK(config_applied), NULL);
	
	gtk_container_add(GTK_CONTAINER(ret), button_hbox);
	
	return(ret);
}
