//standard
#include <cstdlib>
#include <cstdio>	//too used to printf
#include <vector>
#include <cmath>

//system
#include <unistd.h>
#include <sys/types.h>	//for getpid: http://linux.die.net/man/2/getpid
#include <libgen.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>

//gtk
#include <gtk/gtk.h>
#include <gtk/gtkimagemenuitem.h>

//gvfs/gio
#include <gio/gio.h>

//xml
#include <tinyxml/tinyxml.h>

//regex
#include <regexx.hh>
using namespace regexx;

//set the output functions
#ifdef DEBUG
	//use printf for 
	#define DBG_PRINT		printf
#else
	//gobble the debug up if DEBUG isn't set
	#define	DBG_PRINT(...)
#endif
//redirect to stderr
#define ERROR_PRINT(...)	fprintf(stderr, __VA_ARGS__)
//the buffer size for the install directory
#define DIR_BUFFERSIZE 		FILENAME_MAX
//use the file monitor as opposed to the volume monitor (see below)
#define USE_FILE_MONITOR	1

//enum icon_type_t {STOCK, ABSOLUTE, RELATIVE};

// --DECLARATIONS
typedef struct portable_app {
	char* run_path; 	//ABSOLUTE path to the executable
	char* name;			//name of the program
	char* coder;		//authors
	char* version;		//version of the program
	char* release_date;	//release date of the version used
	char* desc;			//description
	char* icon_path;	//ABSOLUTE path to the executable
	
	char* tool_tip;		//the culminating tooltip of this app; intended as a cache
} portable_app;

//a vector of apps, since we don't know which app directories are valid or not
typedef std::vector<portable_app*> APP_VEC;

//step 1: get the directory of this binary; given on autorun

/*
	gets the directory in which this program is install
	this isn't called if this is automatically ran, because this program would be copied
	to /tmp so that the removable media isn't busy, making it available for unmounting
	Instead, the autorun script gives you the mount directory.
*/
static char* get_installdir();
/*
	utility function to merge the paths arg1 and arg2, spews out a new path
	arg1 - the parent directory, ending "/" is optional
	arg2 - the file (or can be more levels of subdirectories), to be appended to arg1 for the return value
*/
static char* dir_merge(const char* arg1, const char* arg2);

//step 2: find out if this apps is on the root of a ejectable device...

/*
	find out where this app is mount, and start monitoring that directory for changes
	and unmounting
*/
static void find_mount();

//steps 3-8

/*
	refresh (or initialize) the menu with the list of apps.
	argc and argv - from main, or can be 0 and NULL if you already called this before
	it called gtk_init and gtk_main if this function is called for the first time.
*/
static void refresh_apps(int argc, char **argv);


//step 3: get apps, validate them, etc.

/*
	get the list (vector) of portable_app pointer for the list of valid applications
	(NULL if could not open dirname)
	dirname - the parent directory of the apps (currently, always the directory "l-portable")
*/
static APP_VEC* get_apps(const char* dirname);
/*
	get the data of the application from their meta.xml file 
	(NULL if could not load the file, if it didn't exist or it was an invalid XML file)
	this respects the version number of the meta.xml file
	dirname - the directory of the app
				This is important for setting $(APP_DIR) (see replace_envvars below).
	metaname - the meta.xml file path
*/
static portable_app* get_appdata(const char* dirname, const char* metaname);
/*
	interpret the text content of the element into an absolute path,
	because they follow the same rules ("type" as an attribute, default relative)
	dirname - the directory of the app
	first_child - the first in the list of elements to check, CANNOT BE NULL
	tag - the tag name, CANNOT BE NULL
*/
static char* interpret_path(const char* dirname, TiXmlNode* first_child, const char* tag);
/*
	get the text content of an element (NULL if e is NULL)
	e - the element in question
*/
static char* get_elementcontent(TiXmlElement* e);
/*
	get the element by the tag name (NULL if not available)
	first_child - the first in the list of elements to check, CANNOT BE NULL
	tag - the tag name, CANNOT BE NULL
*/
static TiXmlElement* get_elementbytag(TiXmlNode* first_child, const char* tag);
/*
	get the content of the element with the tag name (NULL if there isn't an element with that tag name)
	first_child - see get_elementbytag: the first in the list of elements to check, CANNOT BE NULL
	tag - see get_elementbytag: the tag name, CANNOT BE NULL
*/
static char* get_contentbytag(TiXmlNode* first_child, const char* tag);
/*
	replace the environmental variables in the path variable with the
	values of the variables.
	NOTE: the environmental variables are in the format of $(variable-name), such as $(PATH), $(HOME), etc.
	$(APP_DIR) is a special variable set by get_appdata (at least in version 2) that
				evaluates to the specific subdirectory of application whose path is being rplaced right now.
	
*/
static char* replace_envvars(char* path);
/*
	replace one environmental variable. see regexx examples
	A pointer to this function is passed to Regexx::Regexx.replacef(),
	which gives this function the environmental variable (ex:  "$(PATH)")
	and this will return the value of the variable.
*/
static std::string replace_envvar(const RegexxMatch& match);

//step 4: make the menu
/*
	make the menu; the only thing refreshed in refresh_apps
	apps - the list (vector) of apps
*/
static GtkWidget* make_menu(APP_VEC* apps);
/*
	callback, when the menu item clicks; it runs the command
	user_data - a string containing the command, set during menu initialization
*/
static void menu_item_clicked(gpointer user_data);
/*
	make and show the about dialog; called upon when the "About" 
	menu item is clicked; disposed after use
*/
static void show_about_dialog();
//step 5: make the tray icon
/*
	make that tray icon up there
	icon_path - the path of the icon used
*/
static GtkStatusIcon* make_tray_icon(char* icon_path);

/*
	eject/unmount (don't know the difference, don't care) the usb drive,
	if possible
*/
static void portable_eject();
/*
	cleans up and exits the exit code as specified
	exit_code - the return code of the program
*/
static void portable_quit(int exit_code);

//use filemonitor (which can monitor changes to the directory)
#ifdef USE_FILE_MONITOR
	/*
	 	callback for "changed"
	*/
	static void dir_changed(GFileMonitor* monitor, GFile* file
				, GFile* other_file, GFileMonitorEvent event_type
				, gpointer user_data);
//use volumemonitor instead;
#else
	/*
	 	callback for "volume-remove"
	*/
	static void volume_removed(GVolumeMonitor* monitor, GVolume* vol, gpointer user_data);
#endif

//the directory in which this program is installed in
static char install_dir[DIR_BUFFERSIZE];
//install_dir + "l-portable" subfolder, where all the program subfolders are stored
static char* app_dir_base;
//the descriptor of the usb mount, if any (NULL = not on removable media)
static GMount* mount_loc = NULL;
/*
	the monitor of removable volumes (by this user),
	also used to get a list of eligable GMounts
*/
static GVolumeMonitor* monitor = NULL;
//the menu
static GtkWidget* menu = NULL;

#ifdef USE_FILE_MONITOR
	//the filemonitor itself
	static GFileMonitor* mount_monitor = NULL;
	//to thread-safely cancel the monitor
	static GCancellable* monitor_cancel = NULL;
#endif

// --DEFINITIONS
int main(int argc, char **argv)
{
	//just curious what the max size is
	DBG_PRINT("DIR_BUFFERSIZE: %i\n", DIR_BUFFERSIZE);
	
	//step 1: get the directory of this binary; given on autorun
	//given (autorun)
	if(argc > 1)	strcpy(install_dir, argv[1]);
	//not given (debug)
	else			strcpy(install_dir, get_installdir());
	//debug
	DBG_PRINT("DIRNAME: %s\n", install_dir);
	//find this out...
	//app_dir_base = dir_merge(install_dir, "l-portable");
	app_dir_base = install_dir;
	//step 2: find out if this apps is on the root of a ejectable device...
	find_mount();
	//steps 3-7
	refresh_apps(argc, argv);
	//never reached
	return(0);
}

void refresh_apps(int argc, char **argv)
{
	//have I been here before?
	static int gtk_initialized = 0;
	//step 3: get apps, validate them, etc.
	DBG_PRINT("APP_DIR_BASE: %s\n", app_dir_base);
	//get the list of apps
	APP_VEC* apps = get_apps(app_dir_base);
	//was the l-portable/ directory valid?
	if(apps)
	{
		//don't initalize again
		if(!gtk_initialized)	//step 3: initialize
			gtk_init(&argc, &argv);
		//step 4: make the menu
		menu = make_menu(apps);
		// on't need the list anymore
		delete apps;
		if(!gtk_initialized)	
		{
			//step 5: make the tray icon (it automatically displays, so no need for storage)
			make_tray_icon(dir_merge(install_dir, "icon.svg"));
			//step 6: start gtk
			DBG_PRINT("starting gtk...\n");
			//make sure to set this so I don't do this again
			gtk_initialized = 1;
			//doesn't return
			gtk_main();
		}
	}
	//no? =(
	else
	{
		//don't print duplicate error messages...
		//ERROR_PRINT("No application directory (%s) found!\n", app_dir_base);
		//quit, because there are no apps to run
		portable_quit(1);
	}
}

// STEP 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
char* get_installdir()
{
	//step 1: get the directory of this binary
	//you can do that on the command line by: readlink /proc/$PID/exe
	pid_t pid = getpid();
	/*
	 	print it into a string
		since pid is an int (in gcc):
			length = strlen("/proc//exe") (the other stuff) 
						+ log(2^(sizeof(int) * 8)) (the number of digits)
						+ 1 (for null char)
			length = 10 + sizeof(int) * 8 * log(2) + 1
			length = 11 + 2.408 * sizeof(int)
		ceiling function it just to be sure...
	*/
	//just curious...
	DBG_PRINT("sizeof int: %i\n", sizeof(int));
	//debug print out
	const int pid_str_len = (int)ceil(11 + (int)ceil(2.408 * sizeof(int)));
	DBG_PRINT("sizeof pid_str: %i\n", pid_str_len);
	char pid_str[pid_str_len];
	//and finally print it into the string
	sprintf(pid_str, "/proc/%d/exe", pid);
	//print out your results
	DBG_PRINT("PID_STR: %s\n", pid_str);
	//copied from: https://www.securecoding.cert.org/confluence/display/seccode/POS30-C.+Use+the+readlink()+function+properly
	char buf[DIR_BUFFERSIZE];
	//actually do the readlink, and return the length
	ssize_t len = readlink(pid_str, buf, sizeof(buf) - 1);
	//if it actully followed the link
	if(len != -1)
	{
		//insert the null char
		buf[len] = '\0';
		//debug print
		DBG_PRINT("BINARY LOCATION: %s\n", buf);
		//return the directory the program is in
		return(dirname(buf));
	}
	else
	{
		//complain
		ERROR_PRINT("Couldn't get the installation directory!\n");
		//dir
		portable_quit(2);
		//never reached
		return(NULL);
	}
}	/* get_appdir */

// STEP 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
void find_mount()
{
	//initialize so that we can use GLib and GIO
	g_type_init();
	//get the monitor so that...
	monitor = g_volume_monitor_get();
	//..we can get the list of mounts
	GList* first = g_volume_monitor_get_mounts(monitor);
	//it's a (doubly-)linked list, so use another variable to circulate
	GList* current = first;
	//while we haven't run out of mounts and we haven't found one yet
	while(current != NULL && mount_loc == NULL)
	{
		//debug print
		DBG_PRINT("mount: %s at %s\n", g_mount_get_name(G_MOUNT(current->data))
				, g_file_get_path(g_mount_get_root(G_MOUNT(current->data))));
		//get the parent directory, since l-portable-bin is in l-portable/ now
		char* base = new char[strlen(install_dir) + 1];
		//copy over, since dirname modifies the string
		strcpy(base, install_dir);
		//dirname it
		base = dirname(base);
		//is the mount the same as my install directory's parent directory?
		//if yes, set up the mount location and monitors
		if(!strcmp(g_file_get_path(g_mount_get_root(G_MOUNT(current->data))), base))
		{
			//get the mount location
			mount_loc = G_MOUNT(current->data);
			//debug print
			DBG_PRINT("mounted: %s\n", g_mount_get_name(mount_loc));
#ifdef USE_FILE_MONITOR
			//the semaphore (?) for cancelling (it's thread-safe) 
			monitor_cancel = g_cancellable_new();
			/*
				notice that this monitors the APP_DIR, not the install dir,
				because this is more focused...
			*/
			mount_monitor = g_file_monitor_directory(g_file_new_for_path(app_dir_base)
							, G_FILE_MONITOR_WATCH_MOUNTS, monitor_cancel, NULL);
			//connect the signal
			g_signal_connect_swapped(G_OBJECT(mount_monitor), "changed", G_CALLBACK(dir_changed), NULL);
#else
			//otherwise just use the volume monitor we got to list the mounts
			g_signal_connect_swapped(G_OBJECT(monitor), "volume-removed", G_CALLBACK(volume_removed), NULL);
#endif
		}
		//if not, look for the next one;
		else
			current = current->next;
		//clean up
		delete[] base;
	}
}

// STEP 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
//merge as directory and file
char* dir_merge(const char* arg1, const char* arg2)
{
	char* ret = new char[strlen(arg1) + strlen(arg2) + 2];
	char* slash = strrchr(arg1, '/');	//DO NOT DELETE
	if(slash && slash == arg1 + strlen(arg1) - 1)
		strcat(strcpy(ret, arg1), arg2);
	else
		strcat(strcat(strcpy(ret, arg1),  "/"), arg2);
	return(ret);
}

//does the file exist?
int file_exists(const char* fname)
{
	//using open and fstat
	int fd = open(fname, O_RDONLY);
	struct stat *file_stat = new(struct stat);
	int exists = !fstat(fd, file_stat);
	close(fd);
	delete file_stat;
	return(exists);
}

//is this a directory?
int is_dir(const char* fname)
{
	//using open and fstat
	int fd = open(fname, O_RDONLY);
	struct stat *file_stat = new(struct stat);
	int exists = !fstat(fd, file_stat);	//exists?
	close(fd);
	if(exists)
	{
		int ret = S_ISDIR(file_stat->st_mode);
		delete file_stat;
		return(ret);
	}
	delete file_stat;
	return(0);
}

APP_VEC* get_apps(const char* dirname)
{
	//if the directory exists
	if(is_dir(dirname))
	{
		//make the new vec
		APP_VEC* ret = new APP_VEC();
		//get the directory to list the files
		DIR* dir = opendir(dirname);
		//get the structure in the files would be read
		dirent* each_ent;
		//the loop variables: absolute path of the app, meta.xml path, and app data structure
		char* each_name;
		char* each_meta;
		portable_app* each_app;
		//go one at a time...
		while((each_ent = readdir(dir)) != NULL)
		{
			//the directory isn't "." or ".."
			if(strcmp(each_ent->d_name, ".") && strcmp(each_ent->d_name, ".."))
			{
				//find the absolute path of the folder
				each_name = dir_merge(dirname, each_ent->d_name);
				DBG_PRINT("each_name: %s\n", each_name);
				//is it a directory?
				if(is_dir(each_name))
				{
					//yes, move on...
					DBG_PRINT("\tis_dir\n");
					//find the meta.xml file's absolute path
					each_meta = dir_merge(each_name, "meta.xml");
					DBG_PRINT("\tmeta.xml: %s\n", each_meta);
					//does it exist?
					if(file_exists(each_meta))
					{
						//yes, get the data...
						each_app = get_appdata(each_name, each_meta);
						//was there valid data?
						if(each_app)
						{
							//yes, and it to the list
							ret->push_back(each_app);
						}
					}
					//clean up even if the meta.xml file didn't exist
					delete each_meta;
				}
				//clean up even if it wasn't a folder
				delete each_name;
			}
		}
		//finally, close it; otherwise, I can unmount the media
		closedir(dir);
		return(ret);
	}
	else
	{
		//no app directory...
		ERROR_PRINT("Couldn't find application directory (%s)!\n", app_dir_base);
		//this is just a utility function, so don't quit
		return(NULL);
	}
}

char* interpret_path(const char* dirname, TiXmlNode* first_child, const char* tag)
{
	//get the element, because we need it for path type and content
	TiXmlElement* e = get_elementbytag(first_child, tag);
	//get content
	char* content = get_elementcontent(e);
	//is there even an element/content?
	if(content)
	{
		//yes, keep going
		//if it explicably says it's absolute
		if(e->Attribute("type") && !strcmp(e->Attribute("type"), "absolute"))
			//set it as absolute
			return(replace_envvars(content));
		else
			//else, still prepend the app directory it.
			return(dir_merge(dirname, replace_envvars(content)));
		//because it's copied and a\llocated dynamically
		delete content;
	}
	//no, just return NULL
	return(NULL);
}

portable_app* get_appdata(const char* dirname, const char* metaname)
{
	//construct the document
	TiXmlDocument* doc = new TiXmlDocument(metaname);
	//try to load it
	if(doc->LoadFile())
	{
		//yes, get the root (<app>) element
		TiXmlElement* root = doc->RootElement();
		
		//what version is the meta.xml file?
		int meta_version;
		//try top get the version...
		if(root->QueryIntAttribute("version", &meta_version) == TIXML_SUCCESS)
		{
			//yes, debug print
			DBG_PRINT("\tmeta.xml version: %d\n", meta_version);
			//find out how will get the data from the meta.xml file...
			switch(meta_version)
			{
				//current version
				case 2:
				{
					//initialize the structure
					portable_app* ret = new(portable_app);
					//get the first child for convienence, as all of the get***bytag functions need them
					TiXmlNode* first_child = root->FirstChild();
					//construct the tooltip's temporary buffer as a std::string for easy appending
					std::string* tool_tip = new std::string();
					
					//get name
					ret->name = get_contentbytag(first_child, "name");
					DBG_PRINT("\tname: %s\n", ret->name);
					//get authors
					ret->coder = get_contentbytag(first_child, "coder");
					//add to tooltip
					tool_tip->append("<big>Coder</big>: ");
					tool_tip->append(ret->coder);
					//get the version
					ret->version = get_contentbytag(first_child, "version");
					//add to tooltip (notice the newline)
					tool_tip->append("\n<big>Version</big>: ");
					tool_tip->append(ret->version);
					
					//set $(APP_DIR) to the specific directory
					setenv("APP_DIR", dirname, TRUE);
					//get the run path
					ret->run_path = interpret_path(dirname, first_child, "run");
					//get the icon path
					ret->icon_path = interpret_path(dirname, first_child, "icon");
					
					//get the release date
					ret->release_date = get_contentbytag(first_child, "release_date");
					//add to tooltip
					tool_tip->append("\n<big>Release Date</big>: ");
					tool_tip->append(ret->release_date);
					//get the descripting
					ret->desc = get_contentbytag(first_child, "description");
					//add to tooltip
					tool_tip->append("\n<big>Description</big>: ");
					tool_tip->append(ret->desc);
					
					//finally put it in the real structure
					ret->tool_tip = new char[tool_tip->length() + 1];
					strcpy(ret->tool_tip, const_cast<char*>(tool_tip->c_str()));
					//debug print
					DBG_PRINT("\ttool_tip: <>%s<>\n", ret->tool_tip);
					
					//mop up
					delete tool_tip;
					delete doc;
					return(ret);
				}
			}
		}
	}
	//even though it didn't load
	delete doc;
	return(NULL);
}

char* get_elementcontent(TiXmlElement* e)
{
	//does the element even exist?
	if(e)
	{
		//get the text
		const char* text = e->GetText();
		//but we need to copy it (since we're deleting the document thereafter), so check if it's NULL first
		if(text)
		{
			//then copy it over
			char* ret = new char[strlen(text) + 1];
			strcpy(ret, text);
			return(ret);
		}
	}
	return(NULL);
}

TiXmlElement* get_elementbytag(TiXmlNode* first_child, const char* tag)
{
	//while there are more nodes to cycle thru
	while(first_child)
	{
		//if it is an element
		if(first_child->Type() == TiXmlNode::ELEMENT)
		{
			//convert it
			TiXmlElement* e = first_child->ToElement();
			//and check its tag
			if(!strcmp(e->Value(), tag))
			{
				//yes, then return
				return(e);
			}
			//return();
		}
		//cycle thru
		first_child = first_child->NextSibling();
	}
	//no element with the tag found
	return(NULL);
}

char* get_contentbytag(TiXmlNode* first_child, const char* tag)
{
	//just combine those two functions above
	return(get_elementcontent(get_elementbytag(first_child, tag)));
}

char* replace_envvars(char* path) //replace ALL vars
{
	//used for this specific function
	static Regexx r1;
	//convert it for use...
	std::string* path_str = new std::string(path);
	/*
	 	hehe, love how simple this is
	 	basically, everything Regexx encounter something like "$(<anything>)",
	 	it will ask replace_envvar what to place it,
	 	and replace_envvar just gives it the value of the environmental variable
	*/
	DBG_PRINT("rep: %s\n", r1.replace(*path_str, "\\$\\(.*?\\)", "%0").c_str());
	char* ret = const_cast<char*>(r1.replacef(*path_str, "\\$\\(.*?\\)", replace_envvar, Regexx::global).c_str());
	//delete the temporary
	delete path_str;
	return(ret);
}

std::string replace_envvar(const RegexxMatch& match)
{
	//used for this specific function
	static Regexx r2;
	//see above
	return(*(new std::string(getenv(r2.replace(match, "\\$\\((.*?)\\)", "%0").c_str()))));
}

//Step 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4

//adapted from: http://library.gnome.org/devel/gtk-tutorial/stable/c1501.html
GtkWidget* make_menu(APP_VEC* apps)
{
	//new menu...
	GtkWidget* ret = gtk_menu_new();
	//the iterator
    APP_VEC::iterator it = apps->begin();
    //loop variables: the menu item, and the individual icons
    GtkWidget* each_item;
    GtkWidget* each_icon;
    //for each app
    while(it != apps->end())
    {
    	//debug print
    	DBG_PRINT("name: %s\n", (*it)->name);
    	//does meta.xml give you a icon_path and is it 
    	if((*it)->icon_path && file_exists((*it)->icon_path))
    	{
			each_item = gtk_image_menu_item_new_with_label((*it)->name);
			each_icon = gtk_image_new_from_file((*it)->icon_path);
			//set the icon
			gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(each_item), each_icon);
		}
		else
			each_item = gtk_menu_item_new_with_label((*it)->name);
		gtk_widget_set_tooltip_markup(each_item, (*it)->tool_tip);
    	gtk_menu_shell_append(GTK_MENU_SHELL(ret), each_item);
    	g_signal_connect_swapped(G_OBJECT(each_item), "activate", G_CALLBACK(menu_item_clicked), (gpointer)((*it)->run_path));
    	gtk_widget_show(each_item);
    	it++;
	}
	//add a separator
	GtkWidget* separator = gtk_separator_menu_item_new();
	gtk_menu_shell_append(GTK_MENU_SHELL(ret), separator);
	gtk_widget_show(separator);
	//add an about menu item
	GtkWidget* about = gtk_image_menu_item_new_from_stock(GTK_STOCK_ABOUT, NULL);
	gtk_menu_shell_append(GTK_MENU_SHELL(ret), about);
	g_signal_connect_swapped(G_OBJECT(about), "activate", G_CALLBACK(show_about_dialog), NULL);
	gtk_widget_show(about);
	//add an option to install on other devices
	char* installer_path = dir_merge(install_dir, "installer");
	char* command = new char[strlen(installer_path) + 3];
	sprintf(command, "%s &", installer_path);
	delete[] installer_path;
	GtkWidget* install = gtk_image_menu_item_new_with_label("Install On Other Devices");
	gtk_menu_shell_append(GTK_MENU_SHELL(ret), install);
	g_signal_connect_swapped(G_OBJECT(install), "activate", G_CALLBACK(system), (gpointer)command);
	gtk_widget_show(install);
	
	
	//one more seperator
	separator = gtk_separator_menu_item_new();
	gtk_menu_shell_append(GTK_MENU_SHELL(ret), separator);
	gtk_widget_show(separator);
	//quit menu item
	GtkWidget* quit = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, NULL);
	gtk_menu_shell_append(GTK_MENU_SHELL(ret), quit);
	g_signal_connect_swapped(G_OBJECT(quit), "activate", G_CALLBACK(portable_quit), NULL);
	gtk_widget_show(quit);
	//if this is mounted
	if(mount_loc)
	{
		//also add a eject functionality
		GtkWidget* eject = gtk_image_menu_item_new_from_stock(GTK_STOCK_DISCONNECT, NULL);
		gtk_menu_shell_append(GTK_MENU_SHELL(ret), eject);
		g_signal_connect_swapped(G_OBJECT(eject), "activate", G_CALLBACK(portable_eject), NULL);
		gtk_widget_show(eject);
	}
	return(ret);
}

//Step 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5

//adapted from here: http://yettocome.blogspot.com/2007/08/gtk-system-tray-icon-example.html
void tray_icon_on_click(GtkStatusIcon *status_icon, gpointer user_data)
{
	printf("Clicked on tray icon\n");
	//pop the menu up anyways; 3 is the right mouse button, btw.
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 3, 0);
}

void tray_icon_on_menu(GtkStatusIcon *status_icon, guint button, guint activate_time, gpointer user_data)
{
	printf("Popup menu\n");
	//pop the menu up
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, button, activate_time);
}

GtkStatusIcon* make_tray_icon(char* icon_path)
{
	//just curious...
	DBG_PRINT("icon_path: %s\n", icon_path);
	
	GtkStatusIcon *tray_icon;
	//make it use the icon...
	tray_icon = gtk_status_icon_new_from_file(icon_path);
	//connect the signals
	g_signal_connect(G_OBJECT(tray_icon), "activate",	G_CALLBACK(tray_icon_on_click),	NULL);
	g_signal_connect(G_OBJECT(tray_icon), "popup-menu",	G_CALLBACK(tray_icon_on_menu),	NULL);
	//set tooltip
	gtk_status_icon_set_tooltip(tray_icon, "l-portable");
	//show it
	gtk_status_icon_set_visible(tray_icon, TRUE);
	
	//doesn't really matter, not used anyways
	return(tray_icon);
}

void show_about_dialog()
{
	//get a new dialog
	GtkAboutDialog* about_dialog = GTK_ABOUT_DIALOG(gtk_about_dialog_new());
	//set all that jazz
	gtk_about_dialog_set_name(about_dialog, "About l-portable");
	gtk_about_dialog_set_program_name(about_dialog, "l-portable");
	gtk_about_dialog_set_version(about_dialog, "0.0.1");
	gtk_about_dialog_set_copyright(about_dialog, "Copyright (C) 2008 Jeff Chien");
	gtk_about_dialog_set_comments(about_dialog, "Portable Apps for a Portable OS.");	//need better comments
	gtk_about_dialog_set_license(about_dialog, "Gnu GPL version 3");
	gtk_about_dialog_set_website(about_dialog, "http://jchien.awardspace.com/");
	gtk_about_dialog_set_website_label(about_dialog, "jshell");
	const gchar *me[] =
	{
		"Jeff Chien",
		NULL
	}; 
	gtk_about_dialog_set_authors(about_dialog, me);
	gtk_about_dialog_set_artists(about_dialog, me);
	//set the logo (USE ABSOLUTE PATH)
	gtk_about_dialog_set_logo(about_dialog, gdk_pixbuf_new_from_file(dir_merge(install_dir, "icon-medium.png"), NULL));
	//run it
	gtk_dialog_run(GTK_DIALOG(about_dialog));
	//destroy it, or the dialog won't close when the user clicks "Close"
	gtk_widget_destroy(GTK_WIDGET(about_dialog));
}

void menu_item_clicked(gpointer user_data)
{
	//need to append "&" to the command so that it isn't blocking
	char* command = new char[strlen((char*)user_data) + 3];
	//find out our real command
	sprintf(command, "%s &", (char*)user_data);
	//debug print
	DBG_PRINT("execute: %s\n", command);
	//run it
	system(command);
	//don't need it anymore
	delete command;
}

void portable_eject()
{
#ifdef USE_FILE_MONITOR
	/*
		cancel the monitor in advance to clean things up
		I originally thought that not cancelling this montior is why
		I couldn't eject, but it ended up being that I did not close the
		directory back in get_apps.
	*/
	g_cancellable_cancel(monitor_cancel);
#endif
	//umount by force
	g_mount_unmount(mount_loc, G_MOUNT_UNMOUNT_FORCE, NULL, NULL, NULL);
	//and then quit
	portable_quit(0);
}

void portable_quit(int exit_code)
{
	exit(exit_code);
}

#ifdef USE_FILE_MONITOR
void dir_changed(GFileMonitor* monitor, GFile* file, GFile* other_file
				, GFileMonitorEvent event_type, gpointer user_data)
{
	//make sure that this isn't cancelled yet and ignored
	if(!g_cancellable_is_cancelled(monitor_cancel))
	{
		DBG_PRINT("dir_changed\n");
		//ignore the pre-umount even for now
		if(event_type != G_FILE_MONITOR_EVENT_PRE_UNMOUNT)
		{
			//if someone forcibly unmount
			if(event_type == G_FILE_MONITOR_EVENT_UNMOUNTED)
			{
				//yes, we're done!
				portable_quit(0);
			}
			else
			{
				//no, refresh the apps and menu
				refresh_apps(0, NULL);
			}
		}
	}
}
#else
void volume_removed(GVolumeMonitor* monitor, GVolume* vol, gpointer user_data)
{
	//if someone forcibly unmounted...
	if(!strcmp(g_volume_get_name(vol), g_mount_get_name(mount_loc)))
	{
		//we're done!
		portable_quit(0);
	}
}
#endif

